#!/usr/bin/python
"""
Update jenkins jobs for smoke testing
Draws heavily from setup-jenkins-utah.py from
lp:~ubuntu-server-iso-testing-dev/+junk/jenkins-qa-lab
"""
# TODO: fix static validation for raring in utah
# TODO: customize local branch directory
# TODO: add support for remote branches
# TODO: get serial console logs


import argparse
import jenkins
import logging
import os
import sys
import tempfile
import subprocess

from distro_info import UbuntuDistroInfo
from jinja2 import Environment, FileSystemLoader

DEV_SERIES = UbuntuDistroInfo().devel()
ARCHES = ['i386', 'amd64']
INSTALLTYPES = ['desktop', 'server', 'mini', 'alternate']
ISODIR = os.path.join('/', 'var', 'cache', 'utah', 'iso')


def get_parser():
    parser = argparse.ArgumentParser(
        description=('Update smoke testing jenkins jobs.'),
        epilog=("For example:\n"
                "\t%(prog)s /path/to/branch -s " + DEV_SERIES +
                " -u user -p pass -j http://jenkins:8080"
                " -i http://10.98.0.1/iso/\n"),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("branches", nargs='+',
                        help="Path(s) to branch(es) to process")
    parser.add_argument("-t", "--type", help="Image type to use"
                        " (if not supplied, guessed from branch name)")
    parser.add_argument("-d", "--debug", action="store_true",
                        help="Enable debugging")
    parser.add_argument("-n", "--dryrun", action="store_true",
                        help="Dry run mode. Don't execute jenkins commands")
    parser.add_argument("-s", "--series", default=DEV_SERIES,
                        help=("series of Ubuntu to download "
                              "(default=%(default)s)"))
    parser.add_argument("-u", "--username",
                        help="username to use when logging into Jenkins")
    parser.add_argument("-p", "--password",
                        help="username to use when logging into Jenkins")
    parser.add_argument("-j", "--jenkins", default="http://10.189.74.2:8080/",
                        help="URL of jenkins instance to configure test in")
    parser.add_argument("--node", default="aldebaran",
                        help=("Jenkins node were the job should be executed "
                              "(default=%(default)s)"))
    parser.add_argument("-i", "--imageurl", default="http://10.98.0.1/iso/",
                        help="URL where images are located")
    return parser


def update_jenkins(instance, dryrun, jobname, config):
    if not instance.job_exists(jobname):
        logging.debug("Creating Jenkins Job %s ", jobname)
        if dryrun:
            return True, ''
        else:
            try:
                instance.create_job(jobname, config)
                return True, ''
            except Exception as err:
                return False, str(err)
    else:
        logging.debug("Reconfiguring Jenkins Job %s ", jobname)
        if dryrun:
            return True, ''
        else:
            try:
                instance.reconfig_job(jobname, config)
                return True, ''
            except Exception as err:
                return False, str(err)


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logging.basicConfig(level=log_level, format='%(levelname)s: %(message)s')

    logging.debug('Attempting to login to jenkins at ' + args.jenkins + '...')
    # Check to see if authentication information provided
    if args.username is not None:
        logging.debug('...with authentication as user ' + args.username)
        instance = jenkins.Jenkins(args.jenkins,
                                   username=args.username,
                                   password=args.password)
    else:
        logging.debug('...without authentication')
        instance = jenkins.Jenkins(args.jenkins)

    logging.info('Series: ' + args.series)
    logging.debug('Image directory: ' + ISODIR)

    # Set jinja2 template environment
    environment = Environment(loader=FileSystemLoader('templates'),
                              trim_blocks=True)

    failures = []
    for branch_url in args.branches:
        if os.path.isdir(branch_url):
            tempdir = branch_url
        else:
            tempdir = tempfile.mkdtemp(prefix='{}_'.format(parser.prog))
            subprocess.check_call(['bzr', 'branch', '--use-existing-dir',
                                   branch_url, tempdir])
        branchname = os.path.basename(branch_url.rstrip('/'))
        installtype = None
        if args.type is None:
            logging.debug('Guessing installtype from branch')
            for checktype in INSTALLTYPES:
                if checktype in branchname:
                    installtype = checktype
                    break
        else:
            logging.debug('Using passed installtype: ' + args.type)
            installtype = args.type
        if installtype not in INSTALLTYPES:
            logging.error(installtype + ' is not a valid install type')
            logging.error('Skipping branch ' + branch_url)
            continue
        logging.info('Branch: ' + branch_url)
        logging.info('Type: ' + installtype)
        logging.info('Architectures to process: ' + ' '.join(ARCHES))

        commands = {}
        scriptdir = os.path.join(tempdir, 'scripts')
        runlistdir = os.path.join(tempdir, 'runlists')
        branch_isopath = ('{isoroot}/{series}-{installtype}-{{arch}}.iso'
                          .format(isoroot=ISODIR,
                                  series=args.series,
                                  installtype=installtype))
        if os.path.isdir(scriptdir):
            scripts_url = os.path.join(branch_url, 'scripts')
            logging.debug('Processing scripts directory: ' + scriptdir)
            scripts = os.listdir(scriptdir)
            logging.debug('All scripts: ' + ' '.join(scripts))
            for script in list(scripts):
                script_path = os.path.join(scriptdir, script)
                if os.access(script_path, os.X_OK):
                    logging.debug(script + ' is executable, using')
                    scriptname = os.path.splitext(script)[0]
                    command = ('$(./scripts/{test} -i {isopath})'
                               .format(test=script,
                                       isopath=branch_isopath))
                    commands[scriptname] = command
        elif os.path.isdir(runlistdir):
            scripts_url = None
            logging.debug('Processing runlists directory: ' + runlistdir)
            testdir = os.path.join('/', 'var', 'cache', 'utah', branchname)
            logging.debug('Test directory: ' + testdir)
            runlists = os.listdir(runlistdir)
            logging.debug('All runlists: ' + ' '.join(runlists))
            for runlist in list(runlists):
                logging.debug('Processing runlist ' + runlist)
                runlistname = os.path.splitext(runlist)[0]
                if os.path.isfile(os.path.join(tempdir, 'preseeds', runlistname
                                  + '.preseed')):
                    logging.debug('Found preseed for runlist')
                    preseed_url = os.path.join(branch_url, 'preseeds',
                                               runlistname + '.preseed')
                    preseedblock = '-p ' + preseed_url
                elif os.path.isfile(os.path.join(tempdir, 'preseeds',
                                    'default.preseed')):
                    logging.debug('Using default preseed from branch')
                    preseed_url = os.path.join(branch_url, 'preseeds',
                                               'default.preseed')
                    preseedblock = '-p ' + preseed_url
                else:
                    logging.debug('No preseed found, using UTAH default')
                    preseedblock = ''
                runlist_url = os.path.join(branch_url, 'runlists', runlist)
                command = ('run_utah_tests.py {preseedblock} '
                           '-i {isopath} {runlist}'
                           .format(preseedblock=preseedblock,
                                   isopath=branch_isopath,
                                   runlist=runlist_url))
                commands[runlistname] = command
        else:
            logging.error(branch_url + ' does not contain a scripts directory '
                          'or a runlists directory; skipping')
            continue

        for arch in ARCHES:
            logging.info('Processing architecture ' + arch)
            jobs = []
            isopath = branch_isopath.format(arch=arch)
            for name, command in commands.iteritems():
                if name == 'default':
                    continue
                command = command.format(arch=arch)

                jobname = '-'.join((args.series, installtype, arch, 'smoke',
                                    name))
                logging.info('Processing job ' + jobname)

                template = environment.get_template('smoke.xml.jinja')
                smokexml = template.render(name=name,
                                           series=args.series,
                                           installtype=installtype,
                                           arch=arch,
                                           isopath=isopath,
                                           scripts_url=scripts_url,
                                           command=command,
                                           node=args.node)
                result, text = update_jenkins(instance, args.dryrun, jobname,
                                              smokexml)
                if result:
                    jobs.append(jobname)
                else:
                    logging.error(jobname + ' failed to configure: ' + text)
                    failures.append(jobname)
            jobname = '-'.join((args.series, installtype, arch, 'smoke',
                                'default'))
            logging.info('Processing job ' + jobname)
            template = environment.get_template('default.xml.jinja')
            default_command = commands['default'].format(arch=arch)
            defaultxml = template.render(series=args.series,
                                         installtype=installtype,
                                         arch=arch,
                                         isopath=isopath,
                                         jobs=','.join(jobs),
                                         imageurl=args.imageurl,
                                         scripts_url=scripts_url,
                                         command=default_command,
                                         node=args.node)
            result, text = update_jenkins(instance, args.dryrun, jobname,
                                          defaultxml)
            if not result:
                logging.error(jobname + ' failed to configure: ' + text)
                failures.append(jobname)

    if len(failures) == 0:
        logging.info('All jobs updated successfully')
        sys.exit(0)
    else:
        logging.error('These jobs failed to update:')
        for job in failures:
            logging.error(job)
        sys.exit(1)

if __name__ == '__main__':
    main()
